import unreal

EditorUtilitySubsystem = unreal.get_editor_subsystem(unreal.EditorUtilitySubsystem)
EditorUtilitySubsystem.close_tab_by_id("{{ widget_id }}")
#EditorUtilitySubsystem.unregister_tab_by_id("{{ widget_id }}")
